(load "typecheck.scm")
(load "ast.scm")

(define (emit . s)
  (begin
    (apply printf s)
    (display "\n")))

(define (codegen-add xs si env)
  (define (go ys)
    (if (null? ys)
      (emit "movq %rbx, %rax")
      (begin
	(let ((y (car ys)))
	     (if (integer? y)
		 (emit "addq $~a, %rbx" y)
		 (begin
		   (codegen-expr y si env)
		   (emit "addq %rax, %rbx"))))
	(go (cdr ys)))))
  (begin
    (emit "movq $0, %rbx")
    (go xs)))

(define (range s n)
  (if (= 0 n) '()
      (append (range s (- n 1))
	      (list (+ s (- n 1))))))

(define wordsize 8)

(define (codegen-let bindings body si env)
  (let* ((stack-offsets (map (lambda (x) (- si (* x wordsize)))
			     (range 0 (length bindings))))
	 (inner-si (- si (* (length bindings) wordsize)))
	 (names (map car bindings))
	 (exprs (map cadr bindings))
	 (inner-env (append (map cons names stack-offsets) env)))
  (for-each (lambda (expr offset)
	      (codegen-expr expr inner-si env)
	      (emit "movq %rax, ~a(%rsp)" offset))
	    exprs stack-offsets)
  (for-each (lambda (form) (codegen-expr form inner-si inner-env)) body)))

(define (codegen-var name si env)
  (let ((offset (cdr (assoc name env))))
    (emit "movq ~a(%rsp), %rax" offset)))

(define (codegen-expr e si env)
  (cond ((and (list? e) (eq? (car e) '+))
	 (codegen-add (cdr e) si env))
	((let? e) (codegen-let
		   (let-bindings e)
		   (let-body e)
		   si
		   env))
	((var? e) (codegen-var e si env))
	(else (emit "movq $~a, %rax" e))))

(define (codegen program)
  (emit ".text")
  (emit ".p2align 4,,15")
  (emit ".globl _scheme_entry")
  (emit "_scheme_entry:")

  ; handle incoming call from C
  (emit "push %rbp")
  (emit "push %rbx")
  (for-each (lambda (i)
	      (emit (string-append
		     "push %r"
		     (number->string i))))
	    '(12 13 14 15))

  ; our code goes here
  (codegen-expr program 0 '())

  ; restore preserved registers
  (for-each (lambda (i)
	      (emit (string-append
		     "pop %r"
		     (number->string i))))
	    '(15 14 13 12))
  (emit "pop %rbx")
  (emit "pop %rbp")

  (emit "ret"))

(define (compile-to-binary program)
  (when (not (eq? (typecheck program) 'int)) (error #f "not an int"))
  (let ([tmp-path "/tmp/a.s"])
    (when (file-exists? tmp-path) (delete-file tmp-path))
    (with-output-to-file tmp-path
      (lambda () (codegen program)))
    (system "clang -fomit-frame-pointer /tmp/a.s rts.c")))
