(load "ast.scm")
					; ('a, ('b, 'a))
(define (env-lookup env n)
  (if (null? env) (error #f "empty env")			; it's a type equality
      (if (eq? (caar env) n)
	  (cdar env)
	  (env-lookup (cdr env) n))))

(define (env-insert env n t)
  (cons (cons n t) env))

(define abs-arg cadr)

(define cur-tvar 0)
(define (fresh-tvar)
  (begin
    (set! cur-tvar (+ cur-tvar 1))
    (string->symbol
     (string-append "t" (number->string (- cur-tvar 1))))))

(define (last xs)
  (if (null? (cdr xs))
      (car xs)
      (last (cdr xs))))

					
(define (normalize prog) ; (+ a b) -> ((+ a) b)
  (cond
   ((lambda? prog) (list 'lambda (lambda-arg prog) (normalize (lambda-body prog))))
   ((app? prog)
    (if (null? (cddr prog))
	(cons (normalize (car prog)) (normalize (cdr prog))) ; (f a)
	(normalize (cons (cons (car prog) (list (cadr prog))) (cddr prog))))) ; (f a b)
   ((let? prog)
    (append (list 'let
		  (map (lambda (x) (cons (car x) (normalize (cdr x))))
		       (let-bindings prog)))
	    (map normalize (let-body prog))))
   (else prog)))


(define (typecheck prog)
  (define (check env x)
    ;; (display "check: ")
    ;; (display x)
    ;; (display "\n")
    (let
	((res
	  (cond
	   ((integer? x) (list '() 'int))
	   ((boolean? x) (list '() 'bool))
	   ((eq? x 'inc) (list '() '(abs int int)))
	   ((eq? x '+)   (list '() '(abs int (abs int int))))
	   ((symbol? x) (list '() (env-lookup env x)))

	   ((let? x)
	    (let ((new-env (fold-left
			    (lambda (acc bind)
			      (let ((t (check
					(env-insert acc (car bind) (fresh-tvar))
					(cadr bind))))
				(env-insert acc (car bind) (cadr t))))
			    env (let-bindings x))))
	      (check new-env (last (let-body x)))))
		  

	   ((lambda? x)
	    (let* ((new-env (cons (cons (lambda-arg x) (fresh-tvar)) env))
		   (body-type-res (check new-env (lambda-body x)))
		   (subd-env (substitute-env (car body-type-res) new-env)))
	      ;; (display "lambda: ")
	      ;; (display body-type-res)
	      ;; (display "\n")
	      ;; (display subd-env)
	      ;; (display "\n")
	      (list (car body-type-res)
		    (list 'abs
			  (env-lookup subd-env (lambda-arg x))
			  (cadr body-type-res)))))
	   
	   ((app? x) ; (f a)
	    (let* ((arg-type-res (check env (cadr x)))
		   (arg-type (cadr arg-type-res))
		   (func-type-res (check env (car x)))
		   (func-type (cadr func-type-res))
		   
					; f ~ a -> t0
		   (func-c (unify func-type
				  (list 'abs
					arg-type
					(fresh-tvar))))
		   (cs (append func-c (car arg-type-res) (car func-type-res)))
		   
		   (resolved-func-type (substitute cs func-type))
		   (resolved-return-type (caddr resolved-func-type)))
	      ;; (display "app:\n")
	      ;; (display cs)
	      ;; (display "\n")
	      ;; (display func-type)
	      ;; (display "\n")
	      ;; (display resolved-func-type)
	      ;; (display "\n")
	      ;; (display arg-type-res)
	      ;; (display "\n")
	      (if (abs? resolved-func-type)
		  (let ((return-type (substitute cs (caddr resolved-func-type))))
		    (list cs return-type))
		  (error #f "not a function")))))))
      ;; (display "result of ")
      ;; (display x)
      ;; (display ":\n\t")
      ;; (display (cadr res))
      ;; (display "[")
      ;; (display (car res))
      ;; (display "]\n")
      res))
  (cadr (check '() (normalize prog))))


(define (abs? t)
  (and (list? t) (eq? (car t) 'abs)))

(define (tvar? t)
  (and (not (list? t)) (not (concrete? t)) (symbol? t)))

(define (concrete? t)
  (case t
    ('int #t)
    ('bool #t)
    (else #f)))

					; returns a list of pairs of constraints
(define (unify a b)
  (cond ((eq? a b) '())
	((or (tvar? a) (tvar? b)) (~ a b))
	((and (abs? a) (abs? b))
	 (consolidate (unify (cadr a) (cadr b))
		      (unify (caddr a) (caddr b))))
	(else (error #f "could not unify"))))


					; TODO: what's the most appropriate substitution?
					; should all constraints just be limited to a pair?
(define (substitute cs t)
					; gets the first concrete type
					; otherwise returns the last type variable
  (define (get-concrete c)
    (if (null? (cdr c))
	(car c)
	(if (not (tvar? (car c)))
	    (car c)
	    (get-concrete (cdr c)))))
  (fold-left
   (lambda (t c)
     (if (member t c)
	 (get-concrete c)
	 t))
   t cs))

(define (substitute-env cs env)
  (map (lambda (x) (cons (car x) (substitute cs (cdr x)))) env))

(define (~ a b)
  (list (list a b)))

(define (consolidate x y)
  (define (merge a b)
    (cond ((null? a) b)
	  ((null? b) a)
	  (else (if (member (car b) a)
		    (merge a (cdr b))
		    (cons (car b) (merge a (cdr b)))))))
  (define (overlap? a b)
    (if (or (null? a) (null? b))
	#f
	(if (fold-left (lambda (acc v)
			 (or acc (eq? v (car a))))
		       #f b)
	    #t
	    (overlap? (cdr a) b))))

  (cond ((null? y) x)
	((null? x) y)
	(else (let* ((a (car y))
		     (merged (fold-left
			      (lambda (acc b)
				(if acc
				    acc
				    (if (overlap? a b)
					(cons (merge a b) b)
					#f)))
			      #f x))
		     (removed (if merged
				  (filter (lambda (b) (not (eq? b (cdr merged)))) x)
				  x)))
		(if merged
		    (consolidate removed (cons (car merged) (cdr y)))
		    (consolidate (cons a x) (cdr y)))))))
